package edu.georgetown.gucs.dictionary;

import java.io.File;
import java.util.List;
import java.util.Map;
import java.util.HashMap;
import java.util.Map.Entry;
import java.util.concurrent.LinkedBlockingQueue;

import edu.georgetown.gucs.tokenizers.TokenizerList;
import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.SdTextException;
import edu.georgetown.gucs.utility.Token;
import org.slf4j.LoggerFactory;

/**
 * Thread for processing a file to add to a {@linkplain Dictionary}
 *
 * @author Clay Shields
 */
class PartialDictionaryWorker extends Thread {

    private final Dictionary parent;
    private final List<String> tokenizerNames;
    private LinkedBlockingQueue<File> sharedQueue;
    private final File poison;
    private static final org.slf4j.Logger logger = LoggerFactory.getLogger(PartialDictionaryWorker.class);


    /**
     * Constructor that specifies a file to tokenize and add to an existing dictionary
     *
     * @param sharedQueue    the file to tokenize and add to this dictionary
     * @param tokenizerNames the list containing the names of the tokenizers to use on this file
     * @param parent         the dictionary to add the file to
     */
    PartialDictionaryWorker(List<String> tokenizerNames, Dictionary parent, LinkedBlockingQueue<File> sharedQueue,
                            File poison) {
        this.tokenizerNames = tokenizerNames;
        this.sharedQueue = sharedQueue;
        this.parent = parent;
        this.poison = poison;
    }

    /**
     * Adds a file to an existing dictionary using the given tokenizers
     */
    @Override
    public void run() {
        logger.info( "Starting");
        boolean done = false;
        File file = null;

        try {
            // Create a tokenizer list from the names of tokens provided
            TokenizerList tokenizers = new TokenizerList(tokenizerNames);

            while (!done) {

                // process files that come from the parent
                try {
                    file = sharedQueue.take();
                } catch (InterruptedException e) {
                    ExceptionHandler.handleException(e);
                }

                if (file == poison) {
                    done = true;
                    try {
                        sharedQueue.put(poison);
                    } catch (InterruptedException e) {
                        ExceptionHandler.handleException(e);
                    }

                    logger.info("Shutting Down");
                    return;
                }

                logger.info(Thread.currentThread().getName());

                // create a map to hold all the items
                Map<Token, DictionaryEntry> newDictionary = new HashMap<>();


                for (Entry<String, List<Token>> splitterEntry : tokenizers.tokenizeFile(file.getAbsolutePath()).entrySet()) {
                    for (Token token : splitterEntry.getValue()) {
                        if (token.name().length() > 0) {
                            if (newDictionary.containsKey(token)) {
                                newDictionary.get(token).incrementFrequencyCount();
                            } else {
                                DictionaryEntry dictEntry = new DictionaryEntry(token);
                                newDictionary.put(token, dictEntry);
                            }
                        }
                    }

                    if (newDictionary.size() > 0) {

                        this.parent.mergeDictionary(newDictionary);

                    } else {
                        // Global.monitor.printErrorMessage("PartialDictionaryWorker", "No tokens added to dictionary from " + file);
                        logger.info("No tokens added todictionary from {}", file);
                    }
                }
            }
        } catch (SdTextException e) {
            //shut threads off
        }
    }
}