package edu.georgetown.gucs.experiment;

import edu.georgetown.gucs.dictionary.Dictionary;
import edu.georgetown.gucs.fingerprinter.Fingerprint;
import edu.georgetown.gucs.fingerprinter.Fingerprinter;
import edu.georgetown.gucs.matcher.FingerprintMatcher;
import edu.georgetown.gucs.utility.*;
import org.slf4j.LoggerFactory;

import java.io.*;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardOpenOption;
import java.util.*;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * @author Evan B.
 * @since 6/17/14
 */
public class Trial extends Thread {

    private TrialParameters poison;
    private LinkedBlockingQueue<TrialParameters> parameterQueue;
    private Experiment parent;
    private TrialParameters base_parameters;

    private int experimentID;
    private int minScore;

    private Random rnd;

    private String dataset;
    private String matcherName;
    private String fingerprinterName;
    private String mangler;
    private String splitter;

    private Fingerprinter fingerprinter;

    private FileLister fileSource;

    private double dictionarySampleSize;
    private double sampleSize;

    private List<String> tokenizerNames;
    private Pair<Double, Double> idfRange;


    private String fingDirectory = "";
    private String fileName = "";
    private ExperimentResults results;

    private AtomicInteger trialDirNum;


    private static final org.slf4j.Logger logger = LoggerFactory.getLogger(Trial.class);

    /**
     * @param theQueue  all of the parameters
     * @param thePoison indicates when the queue is empty
     * @param parent    the class to pass results
     */
    Trial(LinkedBlockingQueue<TrialParameters> theQueue,
          TrialParameters thePoison,
          Experiment parent) {

        this.parameterQueue = theQueue;
        this.poison = thePoison;
        this.parent = parent;

    }

    /**
     * Runs the trial
     */
    @Override
    public void run() {

        String threadName = Thread.currentThread().getName();

        boolean done = false;
        try {
            while (!done) {
		
                // Grab a set of parameters from the queue
		
                base_parameters = parameterQueue.take();
                System.out.println("[Parameter queue size] " + parameterQueue.size());
                logger.info("taking base_params [queue size] " + parameterQueue.size());

		// make sure it isn't null for some reason
		if (this.base_parameters == null) {
		    logger.info("base_params null [ERROR]");
		    logger.error("Trial {} Thread called with null parameters", threadName );
		    throw new SdTextException("Trial:Run:Trial Thread called with null parameters");
		}
		
		// Quit if we get the poison, and pass it on and exit
		if (base_parameters == poison) {
		    try {
			logger.info("putting poison on queue and quitting");
			logger.info("[Terminating] putting poison on queue");
			System.err.println("Thread got the poison");
			parameterQueue.put(poison);
		    } catch (InterruptedException e) {
			ExceptionHandler.handleException(e);
			logger.info("interruptedException");
		    }
		    logger.info("Throwing a linkedblockingQueue exception");
		    logger.error("Return from poison [kill thread] ");
		    System.err.println("Thread returning after poison");
		    return;
		}

		
		logger.info("[paramQueue size] " + parameterQueue.size());
		
		//initialize object instance fields
		unpackParams();

		logger.info("finished unpacking params:");

		/**
		 * Debug Code
		 * printParams();
		 * if(parameterQueue.size() > 0 ) {
		 * System.out.println(poison.toString());
		 * System.out.println(base_parameters.toString() + "-------------------------\n\n");
		 * }
		 */

		System.out.println("Initializing Results");
		//initialize experimentResults
		results = new ExperimentResults(base_parameters);
		logger.info("[initializing results ");

		results.setThreadName(threadName);
		logger.info("setting threadName " + threadName);

		//  System.out.println("Trial: Getting sublists from fileLister: "
		//        + fileSource.getDirectory());

		Pair<List<File>, List<File>> dictAndSampleLists =
                    fileSource.getSublists(dictionarySampleSize, sampleSize);
		logger.info("DictAndSampleLists A " + dictAndSampleLists.getA().size());
		logger.info("DictAndSampleLists B " + dictAndSampleLists.getB().size());

		List<File> dictionaryList = dictAndSampleLists.getA();
		List<File> sampleList = dictAndSampleLists.getB();

		logger.info("Starting to create Dictionary");
		int dictThreads = 4;
		Dictionary dictionary = new Dictionary();
		dictionary.threadingOn(dictThreads);

		dictionary.setTokenizers(tokenizerNames);
		//Global.debug("FileSource Directory", fileSource.getDirectory());
		dictionary.createPartialDictionary(dictionaryList, fileSource.getDirectory());
		logger.info("[FileSource Directory] " + fileSource.getDirectory());

		if ((idfRange.getA() > 0) || (idfRange.getB() < 1)) {
		    dictionary.trimByIDF(idfRange.getA(), idfRange.getB());
		    logger.info("idfRange in bounds");
		}

		int dictionarySize = dictionary.size();
		logger.info("dictionary size " + dictionary.size());

		results.setDictionarySize(dictionarySize);


		fingerprinter = (Fingerprinter) ObjectBuilder.buildObject(fingerprinterName);

		FingerprintMatcher matcher = (FingerprintMatcher) ObjectBuilder.buildObject(matcherName);
		matcher.setMinimumScore(minScore);

		logger.info("entering computeFingerprints");
		computeFingerprints(dictionary, sampleList);

		logger.info("get results [FINISHED COMPUTE FINGERPRINTS");

		parent.getResults(results);
	    }
	} catch(SdTextException | InterruptedException sdtext) {
	    //shutdown threads here
	} catch (Exception e) {
	    //shutdown threads here
	}

	//is this necessary?
	fingDirectory = "";
    }


    /**
     * stores parameters passed from the Experiment class in local variables
     * these variables are used to dictate the settings for each trial
     */
    private void unpackParams() {

	System.err.println("Inside unpack parms");
	experimentID = base_parameters.getExperimentID();

	matcherName = base_parameters.getMatcherName();

	dataset = base_parameters.getDataset();
	sampleSize = base_parameters.getSampleParameter();
	idfRange = base_parameters.getIDFRange();
	minScore = base_parameters.getMinScore();
	dictionarySampleSize = base_parameters.getDictionarySampleParameter();

	fingerprinterName = base_parameters.getFingerprinterName();
	fingerprinter = null;

	fileSource = base_parameters.getFileSource();
	tokenizerNames = base_parameters.getTokenizerNames();

	mangler = base_parameters.getMangler();
	splitter = base_parameters.getSplitter();

	rnd = base_parameters.getRandom();

	experimentID = base_parameters.getExperimentID();

	// Test to make sure the experiment is not terminating
	if (mangler != null) {
	    fingDirectory = idfRange.getA() + "_" + idfRange.getB() + "_" + experimentID;

	    fileName = mangler;
	    //create a directory for each trial
	    Path path = Paths.get(fingDirectory);
	    if (!Files.isDirectory(path) && !Files.exists(path)) {
		try {
		    Files.createDirectory(path);

		} catch (IOException e) {
		    logger.error("Path Already Exists: {} ", path.toString());
		}
	    }
	}
	trialDirNum.getAndIncrement();
    }


    /**
     * If the mangler is null --> compute non fingerprint
     * if the mangler is not null --> compute fingerprint with mangler
     * store unmangled fingerprint
     * compare mangled fingerprint of the same dataSets;
     * --> score
     */

    //mangle and unmangled fingerprints
    private void computeFingerprints(Dictionary dictionary, List<File> fileSet) throws SdTextException {

	try{
	    List<Fingerprint> fingerprintList = new ArrayList<>();

	    FingerprintMatcher matcher = (FingerprintMatcher) ObjectBuilder.buildObject(matcherName);
	    matcher.setMinimumScore(minScore);

	    ArrayList<Integer> scores = new ArrayList<>();

	    for (File file : fileSet) {

		List<Fingerprint> mangled;
		List<Fingerprint> unmangled;

		//Current file being edited
		String filePath = file.getAbsolutePath();

		try {

		    fingerprinter.setDictionary(dictionary);

		    if (!mangler.contains("null")) {
			fingerprinter.setMangler(mangler, dictionary);
			fingerprinter.setManglerRNG(rnd);
			fingerprinter.setSplitter(splitter);
			//Global.debug("Mangler", mangler);
		    }

		    mangled = fingerprinter.computeFingerprint(filePath);
		    mangled.get(0).setFileName(mangled.get(0).getFileName()
					       + "/" + fingerprinter.getMangler());

		    dictionary.getTokenizers().disableMangler();
		    fingerprinter.setDictionary(dictionary);
		    unmangled = fingerprinter.computeFingerprint(filePath);

		    //differentiate between the mangled and unmangled files
		    unmangled.get(0).setFileName(unmangled.get(0).getFileName() + "/" + mangler);

		    // Global.debug("unmangled", unmangled.get(0).getSdhash());
		    // Global.debug("mangled",mangled.get(0).getSdhash());
		    // System.out.println("FilePath " + filePath);

		    if (mangled.size() <= 0) {
			//debug output
			logger.error("File: {}", filePath);
			logger.error("[mangler] {}", mangler);
			logger.error("[Fingerprinter] {}", fingerprinterName);
			logger.error("[unmangled size]  {}", unmangled.size());
			logger.error("[mangled size]  {}", mangled.size());
			logger.error("[tokenizer Names] {}", tokenizerNames);
		    }

		    //ensure fingerprints were computed


		    int score;
		    //  Global.debug("MatcherName", matcher.getMatcherName());

		    fingerprintList.add(unmangled.get(0));
		    fingerprintList.add(mangled.get(0));

		    if (matcher.getMatcherName().contains("SdHashFingerprintMatcher")) {

			//FIXME test functionality to make sure th is works
			score = matcher.getScore(mangled.get(0).getSdhash(), unmangled.get(0).getSdhash());

			//using cosineSimilarityMatcher
		    } else {

			// Global.debug("mangler", mangler);
			score = matcher.getScore(unmangled.get(0).getFingerprint(),
						 mangled.get(0).getFingerprint());

			//*********
			//DEBUG OUTPUT
			StringBuffer buff = new StringBuffer();
			buff.append("[mangled]").append(" ").append(unmangled.get(0).getFileName()).append("|")
			    .append("[unmangled]").append(" ")
			    .append(mangled.get(0).getFileName()).append("|").append(score).append(" ").append(matcher);

			logger.error(buff.toString());
			//*********

		    }
		    //add scores to later average
		    scores.add(score);

		} catch (IndexOutOfBoundsException e) {
		    logger.error("IError", "Out Of bounds");
		    logger.info("array out of bounds index with mangler");
		}



	    } // Terminate computing all the fingerprints

	    // Put checks to make sure this only occurs once
	    serialize(fingDirectory + "_dict", dictionary);
	    serialize(fileName, fingerprintList);

	    //average scores
	    int count = 0;
	    for (Integer i : scores) {
		count += i;

		//get each score individually
		results.setAggregateScores(i);

	    }

	    if (scores.size() > 0) {
		results.setMatcherScore((count / scores.size()));
	    }
	} catch (SdTextException sdtext) {
	    throw sdtext;
	}catch (Exception e) {
	    ExceptionHandler.handleException(e);
	}
    }

    private void serialize(String fileName, Object fingerprint) {

	try (FileOutputStream fos = new FileOutputStream(fingDirectory + "/" + fileName + ".ser");
	     ObjectOutputStream out = new ObjectOutputStream(fos)) {

	    out.writeObject(fingerprint);

	} catch (IOException e) {
	    e.printStackTrace();
	}
    }

    static void appendStringToFile(Path file, String s) throws IOException  {
	try (BufferedWriter out = Files.newBufferedWriter(file, StandardCharsets.UTF_8, StandardOpenOption.APPEND)) {
	    out.append(s);
	    out.newLine();
	}
    }

}
