package edu.georgetown.gucs.experiment;

import java.util.List;
import java.util.Random;

import edu.georgetown.gucs.fingerprinter.Fingerprint;
import org.apache.xerces.impl.dv.util.Base64;

import edu.georgetown.gucs.dictionary.Dictionary;
import edu.georgetown.gucs.utility.Token;
import edu.georgetown.gucs.utility.FileLister;
import edu.georgetown.gucs.utility.Pair;

/**
 * Provides the trial parameters that are needed for results comparisons
 *
 * @author Lindsay Neubauer
 */
class TrialParameters {

    private final String dataset;
    private final double sampleParameter;
    private final double dictionarySampleParameter;
    private final FileLister fileSource;
    private final String fingerprinterName;
    private final String matcherName;
    private final String mangler;
    private final String splitter;
    private final Pair<Double, Double> idfRange;
    private final int minScore;
    private final int experimentID;
    private final List<String> tokenizerNames;


    private Token file_token;
    private byte[] fingerprint;
    private long randomSeed;
    private Random random;
    private final Dictionary dictionary;

    /**
     * Constructor that sets default values for a trial
     * <p>
     * the <code>int</code> value of the experiment id this trial is a part of; set to -1 if insert = false
     *
     * @param dataset                   the string name of the database to insert trial results
     * @param fileSource                A file lister that will provide the sample files for dictionary creation and for testing
     * @param dictionarySampleParameter the <code>double</code> containing the count or percent to use for dictionary creation in this trial;
     *                                  numbers less than 1 are processed as percents, numbers greater than or equal to 1 are processed as counts
     * @param sampleParameter           the <code>double</code> containing the count or percent to use for the file sample size in this trial;
     *                                  numbers less than 1 are processed as percents, numbers greater than or equal to 1 are processed as counts
     * @param fingerprinterName         the fingerprinter to use in this trial
     * @param matcherName               the matcher to use for fingerprint comparisons in this trial
     * @param mangler                   the string containing the mangler(s)
     * @param splitter                  the string containing the splitter
     */
    protected TrialParameters(int experimentID, String dataset,
                              FileLister fileSource, List<String> tokenizers,
                              double dictionarySampleParameter, double sampleParameter,
                              Pair<Double, Double> idfRange, String fingerprinterName, String matcherName,
                              int minScore, String mangler, String splitter) {

        this.dataset = dataset;
        this.fileSource = fileSource;
        this.tokenizerNames = tokenizers;
        this.dictionarySampleParameter = dictionarySampleParameter;
        this.sampleParameter = sampleParameter;
        this.fingerprinterName = fingerprinterName;
        this.matcherName = matcherName;
        this.mangler = mangler;
        this.splitter = splitter;
        this.idfRange = idfRange;
        this.minScore = minScore;
        this.experimentID = experimentID;
        this.dictionary = null;

    }

    /**
     * This exists so we can make an empty trial parameters object to use as a
     * for when the queue of TrialParameter objects to process is done.
     */
    protected TrialParameters() {

        this.dataset = null;
        this.fileSource = null;
        this.dictionarySampleParameter = 0.0;
        this.sampleParameter = 0.0;
        this.dictionary = null;
        this.fingerprinterName = null;
        this.matcherName = null;
        this.mangler = null;
        this.splitter = null;
        this.minScore = 0;
        this.idfRange = null;
        this.tokenizerNames = null;
        experimentID = 0;
    }

    protected TrialParameters(TrialParameters incoming) {

        this.dataset = getDataset();
        this.fileSource = getFileSource();
        this.tokenizerNames = getTokenizerNames();
        this.dictionarySampleParameter = getDictionarySampleParameter();
        this.sampleParameter = getSampleParameter();
        this.dictionary = null; // Don't need to copy this
        this.fingerprinterName = getFingerprinterName();
        this.matcherName = getMatcherName();
        this.mangler = getMangler();
        this.splitter = getSplitter();
        this.idfRange = getIDFRange();
        this.minScore = getMinScore();
        this.experimentID = getExperimentID();

    }


    /**
     * Sets the non-final information for this trial parameters
     *
     * @param fingerprint the token containing file name and its byte information
     * @param fingerprint the byte array containing the fingerprint
     */
    protected void set(Fingerprint fingerprint) {
        this.fingerprint = fingerprint.getFingerprint();
    }

    /**
     * Sets the random seed for the random number generator
     *
     * @param randomSeed the <code>long</code> the initial seed to use for the random number generator in this trial; can be used
     *                   to enable repeatability across trials
     */
    public void setRandomSeed(long randomSeed) {
        this.randomSeed = randomSeed;
        this.random = new Random(randomSeed);
    }

    /**
     * Provides the token containing file information
     *
     * @return the token file
     */
    public Token getFileToken() {
        return this.file_token;
    }

    public Integer getExperimentID() {
        return experimentID;
    }

    /**
     * Provides the fingerprint
     *
     * @return the byte array containing the fingerprint
     */
    public byte[] getFingerprint() {
        return this.fingerprint;
    }

    /**
     * Provides the name of the dataset
     *
     * @return the string name of the database to insert trial results
     */
    public String getDataset() {
        return this.dataset;
    }

    /**
     * Provides the list of files used for the sample
     *
     * @return the list of sample files to use in this trial
     */
    public FileLister getFileSource() {
        return this.fileSource;
    }

    /**
     * Provides the list of tokenizer names to be used
     *
     * @return the list of tokenizer names
     */
    public List<String> getTokenizerNames() {
        return this.tokenizerNames;
    }


    /**
     * Provides the count or percent to use for the trial's sample
     *
     * @return the <code>double</code> containing the count or percent to use for the file sample size in this trial;
     * numbers less than 1 are processed as percents, numbers greater than or equal to 1 are processed as counts
     */
    public double getSampleParameter() {
        return this.sampleParameter;
    }

    /**
     * Provides the count or percent to use for the trial's dictionary
     *
     * @return the <code>double</code> containing the count or percent to use for dictionary creation in this trial;
     * numbers less than 1 are processed as percents, numbers greater than or equal to 1 are processed as counts
     */
    public double getDictionarySampleParameter() {
        return this.dictionarySampleParameter;
    }

    /**
     * Provides the name of the splitter
     *
     * @return the string name of the splitter to use in this trial
     */
    public String getSplitter() {
        return this.splitter;
    }

    /**
     * Provides the name of the manglers
     *
     * @return the string name of the manglers to use in this trial
     */
    public String getMangler() {
        return this.mangler;
    }

    /**
     * Provides the initial seed to use for the random number generator in this trial
     *
     * @return the initial seed for the random number generator in this trial; can be used to enable repeatability across
     * trials
     */
    public long getRandomSeed() {
        return this.randomSeed;
    }


    /**
     * Provides the random number generator
     * <p>
     * the random number generator in this trial; can be used to enable repeatability across trials
     */
    public void setRandom(Random rnd) {
        this.random = rnd;
    }


    /**
     * Provides the random number generator
     *
     * @return the random number generator in this trial; can be used to enable repeatability across trials
     */
    public Random getRandom() {
        return this.random;
    }


    /**
     * Provides the fingerprinter
     *
     * @return the fingerprinter to use in this trial
     */
    public String getFingerprinterName() {
        return this.fingerprinterName;
    }

    /**
     * Provides the matcher
     *
     * @return the matcher to use for fingerprint comparisons in this trial
     */
    public String getMatcherName() {
        return this.matcherName;
    }

    /**
     * Returns the mininmum score for the matcher
     */
    public int getMinScore() {
        return this.minScore;
    }


    /**
     * Provides the idf range of the dictionary
     *
     * @return a Pair with the min and max idf
     */
    public Pair<Double, Double> getIDFRange() {
        return this.idfRange;
    }


    @Override
    public String toString() {
        StringBuilder stringBuffer = new StringBuilder();

        stringBuffer.append(String.format("%s%n", "Random Seed: " + this.randomSeed));
        stringBuffer.append(String.format("%s%n", "Dataset: " + this.dataset));
        stringBuffer.append(String.format("%s%n", "Fingerprinter name:  " + this.fingerprinterName));
        stringBuffer.append(String.format("%s%n", "Matcher: " + this.matcherName + " (" + this.minScore + ")"));
        String dictionarySampleParameterString = (this.dictionarySampleParameter < 1) ? this.dictionarySampleParameter * 1000 + "%" : this.dictionarySampleParameter + " (count)";
        stringBuffer.append(String.format("%s%n", "Dictionary Parameter: " + dictionarySampleParameterString));
    /* stringBuffer.append(String.format("%s%n",
					  "IDF Range: " + this.idfRange.getA() + ", " + this.idfRange.getB())); */
        stringBuffer.append(String.format("%s%n", "Tokenizers:  " + this.tokenizerNames));
        stringBuffer.append(String.format("%s%n", "Mangler: " + this.mangler));
        stringBuffer.append(String.format("%s%n", "Splitter: " + this.splitter));
        if (this.file_token != null) {
            stringBuffer.append(String.format("%s%n", "File: " + this.file_token));
            stringBuffer.append(String.format("%s%n",
                    "File Bytes: " + this.file_token.start_byte() + "-" + this.file_token.end_byte()));
        }
        if (this.fingerprint != null) {
            stringBuffer.append(String.format("%s%n", "Fingerprint: " + Base64.encode(this.fingerprint)));
        }
        return stringBuffer.toString();
    }
}
