package edu.georgetown.gucs.fingerprinter;

import java.util.ArrayList;
import java.util.ConcurrentModificationException;
import java.util.List;
import java.io.File;
import java.util.Map.Entry;
import java.util.logging.Level;

import edu.georgetown.gucs.tokenizers.TokenizerList;
import edu.georgetown.gucs.utility.*;
import org.apache.commons.lang.ObjectUtils;
import org.slf4j.LoggerFactory;

/**
 * Creates a fingerprint that consist of a bit vector representing the presence or absence of terms in a particular
 * dictionary
 *
 * @author Clay Shields
 * @author Lindsay Neubauer
 */
public class BitVectorFingerprinter extends Fingerprinter {
    private static final org.slf4j.Logger logger = LoggerFactory.getLogger(BitVectorFingerprinter.class);

    /**
     * Constructor that generates the fingerprint name, version, unique identifier (GUID), system identifier, and creating
     * program. Sets the fingerprinter output to display all available information.
     */
    public BitVectorFingerprinter() throws SdTextException {
        super();
    }

    /**
     * Constructor that loads a dictionary and its tokenizers. Generates the fingerprint name, version, unique identifier
     * (GUID), system identifier, and creating program. Sets the fingerprinter output to display all available
     * information.
     *
     * @param dictionaryFilename the filename containing the dictionary to use for this fingerprinter
     */
    public BitVectorFingerprinter(String dictionaryFilename) throws SdTextException {
        super(dictionaryFilename);
    }

    /**
     * Loads a previously generated fingerprinter digest from fingerprinter and dictionary XML files. <br>
     * <br>
     * If dictionaryFilename is null, assumes the full dictionary is contained in the fingerprinter XML file or the file
     * containing the dictionary specified in the fingerprinter XML file exists in the current directory. If the full
     * dictionary is contained in the fingerprinter XML file but does not exist in the current directory, a dictionary XML
     * file containing the dictionary is created.
     *
     * @param fingerprintFilename the string name of the XML file containing the fingerprinter's digest
     * @param dictionaryFilename  the filename containing the dictionary to use for this fingerprinter
     */
    public BitVectorFingerprinter(String fingerprintFilename, String dictionaryFilename) throws SdTextException {
        super(fingerprintFilename, dictionaryFilename);
    }

    /**
     * Computes a byte array fingerprint indicating the presence or absence of each token in this dictionary
     *
     * @return a list of fingerprints that have been computed using the configuration settings
     */
    @Override
    public List<Fingerprint> computeFingerprint(String filename) throws SdTextException {

	//System.err.println("Entering generate fingerprint");

	File input = FileChecker.canRead(filename, "BitVectorFingerprinter", "computeFingerprint");

        List<Fingerprint> bitVectorList = new ArrayList<>();
        try {
	    
            //initial fingerprint set  * Tokenize the input
            AntlrBitSet bitvector = new AntlrBitSet(this.dictionary.size());

            //dictionary = initial tokens
            TokenizerList tokenizers = dictionary.getTokenizers();
        /*
         * Tokenize the input
         *
         */
            logger.info("[Begin BVPF tokenization]");

            for (Entry<String, List<Token>> splitterEntry : tokenizers.tokenizeFile(filename).entrySet()) {

                //  System.out.println("SplitterValue " + splitterEntry.getValue().size());
                logger.info("[Current Mangler] " + splitterEntry.getKey() +
                        " [TokenList size] " + splitterEntry.getValue().size());
                if (splitterEntry.getValue().size() > 0) {

                    //Create a list of the tokens that exist after the bitVectorFingerprinter
                    //weeds out tokens that shouldn't exist
                    ArrayList<Token> tempCopy = new ArrayList<>();
                    tempCopy.addAll(splitterEntry.getValue());

                    //compare tokens in segment to dictionary {use bloom filter}
                    for (Token token : tempCopy) {

                        int position = this.dictionary.getPosition(token);
                        //> -1 means the token exists so add it
                        if (position > -1) {
                            bitvector.add(position);
                        }
                    }

                    logger.info("[BitVectorSize] {} ", bitvector.size());
                    int FIRST_TOKEN = 0;
                    List<Token> tokenList = splitterEntry.getValue();
                    bitVectorList.add(new Fingerprint(filename,
                            bitvector.toBytes(), tokenList,
                            dictionary, tokenList.get(FIRST_TOKEN).start_byte(),
                            tokenList.get(tokenList.size() - 1).end_byte()));
                }
            }
            logger.info("[BitVectorList size] {}", bitVectorList.size());
        } catch (SdTextException sdTextE) {
            throw sdTextE;
        } catch (Exception e) {
            ExceptionHandler.handleException(e);
        }
        return bitVectorList;
    }


    /**
     * Computes a byte array fingerprint indicating the presence or absence of each token in this dictionary
     * does not read from a file -- creates fingerprints from a string and the tokenizers passed
     *
     * @param tokenizer the tokenizerList that will be used to tokenize the input
     * @param str       the string that will be converted to a fingerprint
     * @return a list of fingerprints that have been computed using the configuration settings
     */
    public List<Fingerprint> computeFingerprint(TokenizerList tokenizer, String str) throws SdTextException {

        List<Fingerprint> bitVectorList = new ArrayList<>();
        try {
            AntlrBitSet bitvector = new AntlrBitSet(str.length());

            logger.info("[Begin BVPF tokenization]");

            for (Entry<String, List<Token>> splitterEntry : tokenizer.tokenizeString(str).entrySet()) {

                logger.info("[Current Mangler] {} [TokenList size] {} ",
                        splitterEntry.getKey(), splitterEntry.getValue().size());

                if (splitterEntry.getValue().size() > 0) {

                    //Create a list of the tokens that exist after the bitVectorFingerprinter
                    //weeds out tokens that shouldn't exist

                    //compare tokens in segment to dictionary {use bloom filter}
                    for (Token token : splitterEntry.getValue()) {

                        int position = this.dictionary.getPosition(token);

                        //> -1 means the token exists so add it
                        if (position > -1) {
                            bitvector.add(position);
                        }
                    }

                    logger.info("[BitVectorSize] {}", bitvector.size());
                    int FIRST_TOKEN = 0;
                    List<Token> tokenList = splitterEntry.getValue();
                    bitVectorList.add(new Fingerprint(str,
                            bitvector.toBytes(), tokenList,
                            dictionary, tokenList.get(FIRST_TOKEN).start_byte(),
                            tokenList.get(tokenList.size() - 1).end_byte()));
                }
            }
            logger.info("[BitVectorList size" +
                    "] {} ", bitVectorList.size());
        } catch (SdTextException sdTextE) {
            throw sdTextE;
        } catch (Exception e) {
            ExceptionHandler.handleException(e);
        }
        return bitVectorList;
    }



    //set tokens

    /**
     * A string representation of the BitVectorFingerprinter
     *
     * @return the name of the fingerprinter that was used to create the fingerprints
     */
    @Override
    public String toString() {
        return "BitVectorFingerprinter";
    }

}
