package edu.georgetown.gucs.fingerprinter;

import java.io.File;
import java.math.BigInteger;
import java.util.*;

import edu.georgetown.gucs.dictionary.Dictionary;
import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.FileLister;
import edu.georgetown.gucs.utility.ObjectBuilder;
import edu.georgetown.gucs.utility.SdTextException;
import org.slf4j.LoggerFactory;


/**
 * This is a class that allows threaded parallel creation of fingerprints.
 *
 * @author Clay Shields
 */


public class BulkFingerprinter {

    private final int maxThread;
    private final List<Fingerprint> fpList;
    private int finishedThreadCount = 0 ;
    private final ArrayList<Fingerprint> fingerprint = new ArrayList<>();

    private static final org.slf4j.Logger logger = LoggerFactory.getLogger(BulkFingerprinter.class);

    /**
     * A constructor for when you want to limit the number of threads that are run
     *
     * @param fingerprinterName the name of the fingerprinter-derived class to use for fingerprinting
     * @param files             A FileLister object instantiated with the directory containing the files to fingerprint
     * @param dictionary        The dictionary to be used for fingerprinting
     * @param maxThread         limit on number of threads to be used
     */

    public BulkFingerprinter(String fingerprinterName, FileLister files, Dictionary dictionary, int maxThread) {

        this.maxThread = maxThread;
        this.finishedThreadCount = 0;
        fpList = new ArrayList<Fingerprint>();
    }


    /**
     * A constructor for when you want to limit the number of threads that are run
     *
     * @param maxThread limit on number of threads to be used
     */

    public BulkFingerprinter(int maxThread) {

        this.maxThread = maxThread;
        this.finishedThreadCount = 0;
        fpList = new ArrayList<Fingerprint>();
    }

    /**
     * A constructor that will set the number of threads equal to the number of cores.
     *
     * @param fingerprinterName the name of the fingerprinter-derived class to use for fingerprinting
     * @param files             A FileLister object instantiated with the directory containing the files to fingerprint
     * @param dictionary        The dictionary to be used for fingerprinting
     */


    public BulkFingerprinter(String fingerprinterName, FileLister files, Dictionary dictionary) throws SdTextException {

        this.maxThread = Runtime.getRuntime().availableProcessors();
        this.finishedThreadCount = 0;
        fpList = new ArrayList<Fingerprint>();

    }


    /**
     * Allow a thread to return its results. This is used instead of a synchronized
     * data structure for performance reasons - we want to have many threads read at
     * once for matching
     */

    synchronized void putResults(List<Fingerprint> incoming) {

        for (Fingerprint anIncoming : incoming) {
            fpList.add(anIncoming);
        }

        this.finishedThreadCount++;
    }


    protected Dictionary dict;
    protected Fingerprinter fingerprinter;
    protected String fingerprinterName;

    public BulkFingerprinter(Dictionary dict, String fingerprinterName) throws SdTextException {
        maxThread = Runtime.getRuntime().availableProcessors();
        fpList = new ArrayList<>();
        try {

            this.dict = dict;
            fingerprinter = (Fingerprinter) ObjectBuilder.buildObject(fingerprinterName);
            fingerprinter.setDictionary(dict);
            this.fingerprinterName = fingerprinterName;

        } catch (SdTextException sdtext) {
            throw sdtext;
        } catch (Exception e) {
            ExceptionHandler.handleException(e);
        }
    }

    /**
     * Recursively computes all the fingerprints of documents in a given directory as
     * Base64 encoded strings
     *
     * @param base the directory to fingerprint
     * @return the Base64 encoded string containing the fingerprint and if applicable the start and end bytes for a split
     * file
     */
    public List<Fingerprint> computeAllFingerprints(String fingerprinterName, Dictionary dictionary, FileLister base) throws SdTextException {

	try{

	    fpList.addAll(computeAllFingerprintsUnthreaded(fingerprinterName, dictionary, base));

	} catch (SdTextException sdtext) {
            throw sdtext;
        } catch (Exception e) {
	    ExceptionHandler.handleException(e);
        }

	return fpList;

    }

    public void shutdownThreads(ArrayList<BulkFingerprinterWorker> threadShutdown) throws SdTextException {

        logger.warn("ERROR SHUTTING DOWN THREADS");
	//System.err.println("ERROR SHUTTING DOWN THREADS");

        for (BulkFingerprinterWorker shutdown : threadShutdown) {

            shutdown.interrupt();
        }

    }


    public List<Fingerprint> computeAllFingerprintsUnthreaded(String fingerprinterName, Dictionary dictionary, FileLister base) throws SdTextException {

	logger.info("computeAllFingerprints Starting with {} threads", maxThread);
	//System.err.println("Inside computeAllFingerprintsUnthreaded");

	try{
	
	// Wait for the file lister to finish loading files
	while (!base.getLoaded()) {
	    try {
		Thread.sleep(100);
	    } catch (InterruptedException e) {			
		ExceptionHandler.handleException(e);
	    }
	}

	ArrayList<File> allFiles = base.getFileListerList();

	for (File f : allFiles){
	    //System.err.println("Processing" + f);
	    fpList.addAll(fingerprinter.computeFingerprint(f.getPath()));
	}

	} catch  (SdTextException sdtext) {
            throw sdtext;
        } catch (Exception e) {
            ExceptionHandler.handleException(e);
        }
	
	return fpList;
			
    }

}
