package edu.georgetown.gucs.fingerprinter;

import edu.georgetown.gucs.dictionary.Dictionary;
import edu.georgetown.gucs.dictionary.DictionaryEntry;
import edu.georgetown.gucs.utility.AntlrBitSet;
import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.SdTextException;
import edu.georgetown.gucs.utility.Token;
import org.apache.commons.collections.CollectionUtils;

import java.util.*;

/**
 * @author Evan B.
 * @since 4/23/15.
 */
public class FingerprintTest {

    private  Fingerprint fingerprint_;
    private  String fingerprintFileName_;
    private  Dictionary dictionary_;
    private  String fingerprintGUID_;
    private final List<Integer> fingerprintBitLocations_ = new ArrayList<>();

    private Map<Integer, String> tokenPosition = new HashMap<>();
    private SortTokens sort = new SortTokens(tokenPosition);
    private TreeMap<Integer, String> sorted_results = new TreeMap<>(sort);

    /**
     * Retrieves Fingerprint, Dictionary, and checks GUID #
     * Assumes that the fingerprint contains a dictionary
     *
     * @param fingerprintFileName the name of the file to obtain a fingerprint for testing
     */
    public FingerprintTest(String fingerprintFileName) throws SdTextException {

        try {
            fingerprintFileName_ = fingerprintFileName;

            // Fingerprint does not contain a dictionary;
            Fingerprinter fingerprint = new Fingerprinter(fingerprintFileName, null);

            fingerprint_ = fingerprint.getBase64Fingerprints().get(0);
            fingerprintGUID_ = fingerprint.getGUID();

            dictionary_ = fingerprint.getDictionary();
            checkGUID();
        } catch (SdTextException sdTextE) {
            throw sdTextE;
        } catch (Exception e) {
            ExceptionHandler.handleException(e);
        }
    }

    /**
     * Retrieves a Fingerprint, Dictionary, and checks GUID#
     *
     * @param fingerprintFileName the file that contains the fingerprint for testing
     * @param dictionaryFileName  the file name containing the dictionary for a given fingerprint
     */
    public FingerprintTest(String fingerprintFileName,
                           String dictionaryFileName) throws SdTextException {
        try {
            fingerprintFileName_ = fingerprintFileName;

            // Fingerprint does not contain a dictionary;
            Fingerprinter fingerprint = new Fingerprinter(fingerprintFileName, dictionaryFileName);

            fingerprint_ = fingerprint.getBase64Fingerprints().get(0);
            fingerprintGUID_ = fingerprint.getGUID();

            dictionary_ = new Dictionary(dictionaryFileName);
            checkGUID();
        } catch (SdTextException sdTextE) {
            throw sdTextE;
        } catch (Exception e) {
            ExceptionHandler.handleException(e);
        }
    }

    /**
     * old school method for masking bits
     * determine if a bit position contains a value
     */
    public void findBitPositions() {


        int counter = 0;
        for (byte b : fingerprint_.getFingerprint()) {
            for (int mask = 0x01; mask != 0x100; mask <<= 1) {
                boolean value = (b & mask) != 0;
                if (value) {
                    // System.out.println("BitSet@: " + counter);
                    fingerprintBitLocations_.add(counter);
                }
                counter++;
            }
        }
    }


    public TreeMap<Integer, String> bitsMatchingTokens() {

        for (Map.Entry<Token, DictionaryEntry> dictionary : dictionary_.getFullDictionary().entrySet()) {

            tokenPosition.put(dictionary.getValue().getPosition(), dictionary.getKey().toString());

        }

        for (Integer i : fingerprintBitLocations_) {

            sorted_results.put(i, tokenPosition.get(i));

        }

        for (Map.Entry<Integer, String> sort : sorted_results.entrySet()) {

            System.out.println(sort.getKey() + ":" + sort.getValue());

        }
        System.out.println();
        return sorted_results;
    }

    /**
     * @return the fingerprint used to create the test object
     */
    public Fingerprint getFingerprint() {
        return fingerprint_;
    }

    /**
     * @param fingerprintIncoming the fingerprint to compare against
     */
    public void antlrBitSet(Fingerprint fingerprintIncoming, Dictionary dictionary) {


        AntlrBitSet bitSet_1 = new AntlrBitSet(fingerprint_.getFingerprint());
        AntlrBitSet bitSet_2 = new AntlrBitSet(fingerprintIncoming.getFingerprint());

        System.out.println("\n");
        List<String> termsNew = dictionaryTerms(bitSet_2.toString(), dictionary); //incoming
        List<String> termsOld = dictionaryTerms(bitSet_1.toString(), dictionary_); //base

        Collections.sort(termsNew);
        Collections.sort(termsOld);


        ArrayList commonList = (ArrayList) CollectionUtils.removeAll(termsNew, termsOld);

        if (commonList.size() == termsNew.size() || commonList.size() == termsOld.size())
            System.out.println("stayed the same size");
        // System.out.println("Removed Terms: " + removeTerms + " New Terms: " + termsNew.size() + " Old Terms: " + termsOld.size()) ;

        termsNew.removeAll(commonList);
        termsOld.removeAll(commonList);
        System.out.println("Newterms: " + termsNew + " size " + termsNew.size());
        System.out.println("Oldterms: " + termsOld + " size " + termsOld.size());

        printIntersections(termsNew, termsOld);

    }

    /**
     * @param terms     the dictionary tokens that are being compared
     * @param compareTo the incoming list that will be compared
     */
    private void printIntersections(List<String> terms, List<String> compareTo) {

        if (terms.size() > 0) {

            for (String t : terms) {
                compareTo.remove(t);
            }
        }

        if (compareTo.size() == 0) {
            System.out.println("Fingerprints matched");
        }

    }

    public List<String> dictionaryTerms(String str, Dictionary dict) {

        List<String> terms = new ArrayList<>();
        Map<Integer, Token> tempDict = dict.getDictionaryAndPosition();

        for (String s : str.split(",")) {

            Integer position = Integer.parseInt(s);
            if (tempDict.containsKey(position)) {
                // System.out.println("Position: " + position +
                // " Token: " +  tempDict.get(position).toString());

                terms.add(tempDict.get(position).toString());

            }
        }
        return terms;
    }


    /**
     * Takes a GUID # from the fingerprint provided
     * and tests to ensure that the dictionary/fingerprint originated from the same file
     */
    public void checkGUID() {
        System.out.println(fingerprintFileName_);
        if (!checkGUIDHelper()) {
            System.out.println("GUID MisMatch: ");
            System.out.println("\t [dictionaryGUID] " + dictionary_.getGUID());
            System.out.println("\t [fingerprintGUID] " + fingerprintGUID_);
            System.out.println();
        }
    }

    /**
     * @return a boolean value to indicate if the dictionary GUID match
     */
    private boolean checkGUIDHelper() {

        return dictionary_.getGUID().equals(fingerprintGUID_);

    }

    /**
     * @return the filename of the fingerprint
     */
    public String getFileName() {
        return fingerprintFileName_;
    }

    /**
     * @return the fingerprint dictionary
     */
    public Dictionary getDictionary() {
        return dictionary_;
    }

}

class SortTokens implements Comparator<Integer> {

    Map<Integer, String> tokenMap;

    SortTokens(Map<Integer, String> tokenMap) {
        this.tokenMap = tokenMap;
    }

    @Override
    public int compare(Integer o1, Integer o2) {

        return tokenMap.get(o1).compareTo(tokenMap.get(o2));

    }

}