package edu.georgetown.gucs.matcher;

import java.util.SortedMap;
import java.util.List;
import java.util.ArrayList;

import edu.georgetown.gucs.tokenizers.*;
import edu.georgetown.gucs.dictionary.*;
import edu.georgetown.gucs.matcher.*;
import edu.georgetown.gucs.utility.*;
import edu.georgetown.gucs.fingerprinter.*;


import org.slf4j.LoggerFactory;

/**
 * This is a class that allows threaded parallel comparison of fingerprints.
 * 
 * @author Clay Shields
 **/

public class BulkMatcher {

    private String matcherName;
    private Integer min_score;
    private Integer numThreads;
    

    private static final org.slf4j.Logger logger = LoggerFactory.getLogger(BulkMatcher.class);

    private void init(String matcherType, int score, int maxThread){

	this.matcherName = matcherType;
	this.min_score = score;
	this.numThreads = maxThread;

    }
    
    /**
     * A constructor for when you want to limit the number of threads that are run
     * @param matcherName
     *           the name of the fingerprinter-derived class to use for fingerprinting
     * @param results
     *      A map to insert the results in
     * @param minScore
     *      The minimum score required to be added to the results
     * @param maxThread
     *      limit on number of threads to be used
     */
    BulkMatcher(String matcherType, int score, int maxThread){

	init(matcherType, score, maxThread);
	
    }


    /**
     * A constructor that doesn't require a minimum score
     *
     **/
    
    public BulkMatcher(String matcherType, int maxThread) {

	init(matcherType, Global.MIN_SCORE, maxThread);
    }

     
    /**
     * A method to do multithreaded comparisons between a single fingerprint and a list of other fingerprints
     *
     * @param fing
     *         The single fingerprint to compare
     * @param fingerprints
     *           a list of the fingerprint/files pairs to compare the single fingerprint against    
     */
     

    public List<Pair<String, Integer>> compareFingerprintToList(Fingerprint fing, List<Fingerprint> fingerprint) throws SdTextException{

	logger.info("Entry");
	//	System.err.println("Entry into compareFingerprintToList");			  
	
	if (fing == null){
	    throw new SdTextException("BulkMatcher:compareFingerprintToList:null fingerprint");
	}

	//	System.err.println("Check fingerprint");
	if (fingerprint == null){
	    throw new SdTextException("BulkMatcher:compareFingerprintToList:null fingerprint list");
	}
	
	// Clay - we will fix the threading on this. For the time being, use the unthreaded version
	try{
	    //System.err.println("Calling compareFingerprintToListUnthreaded");
	    logger.info("Calling compareFingerprintToListUnthreaded");
	    return compareFingerprintToListUnthreaded(fing, fingerprint);
	}
	catch (SdTextException e){
	    throw e;
	}
    }


    
    /**
     * A method to do n^2 comparisons among all fingeprints present
     *
     * @param fingerprints
     *           a list of the fingerprint/files pairs to compare     
     */
     

    public List<Pair<Pair<String, String>, Integer>> compareAllFingerprints(List<Fingerprint> fingerprints)
	throws SdTextException{
	
	
	return compareAllFingerprintsUnthreaded(fingerprints);
	
    }


    /**
     * A method to do unthreaded comparisons between a single fingerprint and a list of fingerprints
     *
     * @param fing
     *         The single fingerprint to compare
     * @param fingerprints
     *           a list of the fingerprint/files pairs to compare the single fingerprint against    
     */ 

    public ArrayList<Pair<String, Integer>> compareFingerprintToListUnthreaded(Fingerprint fing, List<Fingerprint> fingerprint) throws SdTextException {

        ArrayList<Pair<String, Integer>> arrList = new ArrayList<>();

        try {
            logger.info("FingerprintSize: {}", fingerprint.size());

	    //System.err.println("Builing matcher");
            FingerprintMatcher matcher = (FingerprintMatcher) ObjectBuilder.buildObject(matcherName);

	    if (matcher == null){
		throw new SdTextException("BulkMatcher:compareFingerprintToListUnthreaded:ObjectBuilder returned null matcher");
	    }
	    
	    //	    System.err.println("Processing fingerprints");
            for (Fingerprint fingerprint1 : fingerprint) {

                int score = matcher.getScore(fing.getFingerprint(), fingerprint1.getFingerprint());
		if (score >= min_score){
		    arrList.add(new Pair<>(fingerprint1.getFileName(), score));
		}

            }
        } catch (SdTextException sdtextE) {
            throw sdtextE;
        } catch (Exception e) {
            ExceptionHandler.handleException(e);
        }
        return arrList;
    }


 /**
     * A method to do unthreaded comparisons between a single fingerprint and a list of fingerprints
     *
     * @param fing
     *         The single fingerprint to compare
     * @param fingerprints
     *           a list of the fingerprint/files pairs to compare the single fingerprint against    
     */ 

    public ArrayList<Pair<Pair<String, String>, Integer>> compareAllFingerprintsUnthreaded(List<Fingerprint> fingerprint)
	throws SdTextException {

        ArrayList<Pair<Pair<String, String>, Integer>> arrList = new ArrayList<>();

        try {
	    
            FingerprintMatcher matcher = (FingerprintMatcher) ObjectBuilder.buildObject(matcherName);

	    
	    // Iterate through the files. Files don't need to match themselves, and order
	    // doesn't matter, so it is the lower triangle
	    for (int i = 0; i < fingerprint.size(); i++){
		for (int k = i + 1; k < fingerprint.size(); k++){

		    // Compute the score
		    int score = matcher.getScore(fingerprint.get(i).getFingerprint(), fingerprint.get(k).getFingerprint()); 
		    // Pack up the names
		    Pair<String, String> names = new Pair<String,String>(fingerprint.get(i).getFileName(),
									 fingerprint.get(k).getFileName());
		    // Save the score and the names
		    if (score >= min_score){
			arrList.add(new Pair<>(names, score));
		    }
		}
	    }
        } catch (SdTextException sdtextE) {
            throw sdtextE;
        } catch (Exception e) {
            ExceptionHandler.handleException(e);
        }
        return arrList;
    }

    
}
