package edu.georgetown.gucs.matcher;

import java.io.*;
import java.util.*;
import java.util.Map.Entry;


import edu.georgetown.gucs.dictionary.Dictionary;
import edu.georgetown.gucs.fingerprinter.Fingerprint;
import edu.georgetown.gucs.fingerprinter.Fingerprinter;
import edu.georgetown.gucs.fingerprinter.BulkFingerprinter;
import edu.georgetown.gucs.utility.*;
import org.slf4j.LoggerFactory;


/**
 * Outputs a list of files and their scores, comparing a given fingerprint to a directory of files. Files are organized
 * by scores with higher scores appearing higher in the list. Providing a minimum score and a dictionary (if the
 * fingerprint does not already contain one) is optional.
 *
 * @author Lindsay Neubauer
 */
public class CompareFingerprint  {


    protected FileLister fileLister;

    protected Fingerprinter fingerprint;
    protected FingerprintMatcher matcher;
    protected String matcherName;
    protected int minimum_score;
    protected Dictionary dictionary;
    protected SortedMap<Integer, List<String>> scores;
    protected int maxThread;
    protected ArrayList<Pair<String, Integer>> resultSet;

    private static final org.slf4j.Logger logger = LoggerFactory.getLogger(CompareFingerprint.class);

    /**
     * Comparator that orders scores highest to lowest
     */
    private static class HightoLowScore implements Comparator<Integer>, Serializable {
        private static final long serialVersionUID = 5679507537454070748L;

        public HightoLowScore() {
        }

        @Override
        public int compare(Integer one, Integer two) {
            return two.compareTo(one);
        }
    }

    /**
     * Constructor that sets the matcher, fingerprint file, directory, minimum score and dictionary to use for this
     * comparison
     *
     * @param match           the string name of the matcher to use for this comparison
     * @param fingerprintFile the string filename of the fingerprint to compare with this directory
     * @param directory       the string directory path to compare with this fingerprint
     * @param minimumScore    the <code>int</code> minimum score needed to output a file for this comparison
     * @param dictionary      the string filename of the dictionary to use for this comparison; this dictionary will only be used if the
     *                        fingerprint file does not already contain a full dictionary
     *                        added global variables to set this.minimum_Score -- evan
     */
    public CompareFingerprint(String match, String fingerprintFile, String directory, int minimumScore, String dictionary) throws SdTextException {
        try {

	    //System.err.println("Entering CompareFingerprint");
	    
	    if (match == null){
		throw new SdTextException("CompareFingerprint:CompareFingerprint:null match");
	    }


	    if (minimumScore < Global.MIN_SCORE  || minimumScore > Global.MAX_SCORE){
		throw new SdTextException("CompareFingerprint:CompareFingerprint: minimum score out of range");
	    }
	    else {
		this.minimum_score = minimumScore;
		//System.err.println("Min score of: " + minimum_score);
	    }

	    //System.err.println("Checking fingerprint");
	    FileChecker.canRead(fingerprintFile, "CompareFingerprint", "CompareFingerprint");
	    FileChecker.canRead(directory, "CompareFingerprint", "CompareFingerprint");

	    if (dictionary != null){ 
		//System.err.println("Checking dictionary");
		FileChecker.canRead(dictionary, "CompareFingerprint", "CompareFingerprint");
	    }
	    else {
		//System.err.println("No dictionary provided to check");
	    }
	    
	    //	    System.err.println("Building matcher");
	    
	    this.matcherName = match;
            this.matcher = (FingerprintMatcher) ObjectBuilder.buildObject(match);

	    //System.err.println("Building scores");
            this.scores = Collections.synchronizedSortedMap(new TreeMap<Integer, List<String>>(new HightoLowScore()));

	    //System.err.println("Calling file lister");
            this.fileLister = new FileLister(directory);
            this.fileLister.loadQueue();


	    // if there is no dictionary specified, try and get it from the fingerprint
	    //System.err.println("Get the dictionary...");

	    this.fingerprint = new Fingerprinter(fingerprintFile, dictionary);

	    if (dictionary == null){
		//System.err.println(" ... from fingerprint");
                this.dictionary = this.fingerprint.getDictionary();
		if (this.dictionary == null){
		    throw new SdTextException("CompareFingerprint:CompareFingerprint: Expected to find a dictionary in the fingerprint, but it wasn't there. Try providing a dictionary as a parameter.");
		}
	    }
	    else {
		//System.err.println("... from parameter");		
		this.dictionary = new Dictionary(dictionary);
	    }

	    //System.err.println("Getting thread count");
            this.maxThread = Runtime.getRuntime().availableProcessors();
        } catch (SdTextException sdtextE) {
            throw sdtextE;
        } catch (Exception e) {
            ExceptionHandler.handleException(e);
        }
    }


    /**
     * Compute a fingerprint for each file in this directory and compare it to each other using this matcher; if the
     * matcher returns boolean values, true values are given a score of 100 and false values are given a score of 0
     */

    public List<Pair<String, Integer>> compareFingerprintDirectory() throws SdTextException {

	List<Pair<String, Integer>> unsortedScores;
	
        try {

	    logger.info("Entry");
	    
            List<Fingerprint> results = new ArrayList<>();

            BulkFingerprinter f = new BulkFingerprinter(dictionary, "BitVectorFingerprinter");

            //System.err.println("Fingerprinting directory... " + System.currentTimeMillis());
	    logger.info("Fingerprinting directory");
	    
            results.addAll(f.computeAllFingerprints("BitVectorFingerprinter", dictionary, fileLister));

	    logger.info("Fingerprinting complete");
	    //System.err.println(" ...done " + System.currentTimeMillis());
	    
	    //System.err.println("Matching fingerprints... " + System.currentTimeMillis());
	    logger.info("Creating BulkMatcher with matcherName " + matcherName);

	    BulkMatcher bulkMatch = new BulkMatcher(matcherName, minimum_score, maxThread);
	    
	    if (bulkMatch == null){
		throw new SdTextException("CompareFingerprint:compareFingerprintDirectory:null bulk matcher created");
	    }

   	    if (fingerprint == null){
		throw new SdTextException("CompareFingerprint:compareFingerprintDirectory:fingerprint is null");
	    }
	    
	    if (results == null){
		throw new SdTextException("CompareFingerprint:compareFingerprintDirectory:results are null");
	    }

	    
	    //System.err.println("Call to bulk matcher");
	    logger.info("Calling BulkMatcher");

	    //System.err.println("Fingerprint is: " + fingerprint.getBase64Fingerprints().get(0));
	    logger.info("Fingerprint is: " + fingerprint.getBase64Fingerprints().get(0));
	    //System.err.println("Results is: " + results);
	    //logger.info("Results is: " + results);
	    
	    unsortedScores = bulkMatch.compareFingerprintToList(fingerprint.getBase64Fingerprints().get(0), results);

	    logger.info("Sorting scores");
	    //System.err.println("Sorting scores with min score: " + minimum_score);
	    
	    // Add the returned scores to the local sorted scores variable for printing and such
	    for (Pair<String, Integer> temp : unsortedScores){

		Integer currentScore = temp.getB();
		String currentFile = temp.getA();		

		
		if (currentScore >= minimum_score){
		    List<String> filesForScore = scores.get(currentScore);
		    
		    if (filesForScore == null){
			filesForScore = new ArrayList<String>();
			filesForScore.add(currentFile);
			scores.put(currentScore, filesForScore);
		    }
		    else {		    
			filesForScore.add(currentFile);
		    }
		}
	    }
	    
	    
	    //System.err.println(" ...done " + System.currentTimeMillis());

	    logger.info("Exit");
            return unsortedScores;


	} catch (SdTextException sdtextE) {
            throw sdtextE;
        } catch (Exception e) {
            ExceptionHandler.handleException(e);
        }
	return null;
    }


    /**
     * Provides all the files and their similarity scores; if a minimum score is provided, only output those files with
     * scores greater than or equal to the minimum
     *
     * @return the string containing all the score results
     */
    private String getScores() throws SdTextException {
        StringBuilder buffer = new StringBuilder();
        boolean printedSomething = false;
        try {
            for (int i = Global.MAX_SCORE; i >= this.minimum_score; i--) {
                if (this.scores.containsKey(i)) {
                    for (String score : this.scores.get(i)) {
                        if (i < Global.MAX_SCORE && i > 10) buffer.append(String.format("%s%n", score + "|0" + i));
                        else if (i <= 10) buffer.append(String.format("%s%n", score + "|00" + i));
                        else buffer.append(String.format("%s%n", score + "|" + i));
                    }
                    printedSomething = true;
                }
            }
            if (!printedSomething) {
                buffer.append(String.format("%s",
                        "CompareDictionary: There are no files in the " +
                                "directory with a score higher than " + this.minimum_score));
            }
        } catch (Exception e) {
            ExceptionHandler.handleException(e);
        }
        return buffer.toString();
    }


    /**
     * Print all the files and their similarity scores; if a minimum score is provided, only output those files with
     * scores greater than or equal to the minimum
     */
    public void printScores() throws SdTextException {
        try {
            System.out.println(getScores());
        } catch (SdTextException sdtextE) {
            throw sdtextE;
        } catch (Exception e) {
            ExceptionHandler.handleException(e);
        }
    }


}
