package edu.georgetown.gucs.matcher;

import edu.georgetown.gucs.fingerprinter.BitVectorFingerprinter;
import edu.georgetown.gucs.fingerprinter.Fingerprint;
import edu.georgetown.gucs.utility.AntlrBitSet;
import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.Global;
import edu.georgetown.gucs.utility.SdTextException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Provides a score based on a cosine similarity match to be used with a {@linkplain BitVectorFingerprinter}. Uses the
 * {@linkplain AntlrBitSet} to do the calculations.
 *
 * @author Clay Shields
 */

public class CosineSimilarityFingerprintMatcher extends FingerprintMatcher {

    private static final Logger logger = LoggerFactory.getLogger(CosineSimilarityFingerprintMatcher.class);

    /**
     * Constructor that sets the minimum score to use for matching two fingerprints to zero
     */
    public CosineSimilarityFingerprintMatcher() {
        super();
    }

    /**
     * Constructor that sets the minimum score to use for matching two fingerprints
     *
     * @param score the minimum score needed to be considered a match
     */
    public CosineSimilarityFingerprintMatcher(int score) {
        super(score);
    }

    /**
     * Constructor that sets the minimum score to use for matching two fingerprints. The string parameter is needed for
     * use with ObjectBuilder.
     *
     * @param score the string value of the minimum score needed to be considered a match
     */
    public CosineSimilarityFingerprintMatcher(String score) {
        super(Integer.parseInt(score));
    }

    /**
     * Determines that the two fingerprints are matching if their similarity score is at or above the minimum score for
     * this fingerprinter
     */
    @Override
    public boolean match(byte[] fp1, byte[] fp2) throws SdTextException {
        try {
            return getScore(fp1, fp2) >= this.minimum_score;
        } catch (SdTextException sdtextE) {
            throw sdtextE;
        } catch (Exception e) {
            ExceptionHandler.handleException(e);
        }
        return false;
    }

    /**
     * Determines the cosine similarity score for these two fingerprints
     */
    @Override
    public int getScore(byte[] fp1, byte[] fp2) throws SdTextException {
        try {
            if (fp1 == null || fp2 == null) {
                ExceptionHandler.handleException(new Exception("Null fingerprints given for matching"));
            }

            // Convert the byte arrays to BitSets

            AntlrBitSet matchingBits = new AntlrBitSet(fp1);
            AntlrBitSet secondBitset = new AntlrBitSet(fp2);
            int firstDegree = matchingBits.degree();
            int secondDegree = secondBitset.degree();

            // If either degree of the arrays is zero, return zero
            // @evan swapped 0 for Global.MIN_SCORE
            if (firstDegree == Global.MIN_SCORE || secondDegree == Global.MIN_SCORE) {
                return Global.MIN_SCORE;
            }

            // Find matching bits
            matchingBits.andInPlace(secondBitset);
            return (int) ((matchingBits.degree() / Math.sqrt(firstDegree * secondDegree)) * Global.MAX_SCORE);
        } catch (SdTextException sdtextE) {
            throw sdtextE;
        } catch (Exception e) {
            ExceptionHandler.handleException(e);
        }
        return -1;
    }

    @Override
    public String toString() {
        return "CosineSimilarityFingerprintMatcher";

    }


    public synchronized FingerprintMatcher getScore(Fingerprint fp1, Fingerprint fp2) throws SdTextException {
        try {
            //Global.debug(fp1.getBase64Fingerprint(), fp2.getBase64Fingerprint());
            comparedFile = fp1.getFileName();
            compareFile = fp2.getFileName();
            score = getScore(fp1.getFingerprint(), fp2.getFingerprint());
            return this;

        } catch (SdTextException sdtextE) {
            throw sdtextE;
        } catch (Exception e) {
            ExceptionHandler.handleException(e);
        }
        return this;
    }

}
