package edu.georgetown.gucs.tokenizers;

import java.util.Collections;
import java.util.List;
import java.util.Vector;
import java.util.logging.Level;
import java.util.logging.Logger;

import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.Global;
import edu.georgetown.gucs.utility.SdTextException;
import edu.georgetown.gucs.utility.Token;

/**
 * Eliminates tokens nested inside markup language tags; assumes that tokens have been split by line rather than using
 * whitespace
 *
 * @author Lindsay Neubauer
 */
public class StripMarkupTokenizer extends Tokenizer {

    private static final long serialVersionUID = -1366304768764987607L;
    private boolean keepScript = false;

    private static final Logger logger = Logger.getLogger(Tokenizer.class.getName());

    /**
     * Constructor that specifies whether to keep tokens nested inside script tags the default is to
     * eliminate these tokens
     *
     * @param keepScript the string value (true or false) specifying whether to keep tokens nested inside script
     *                   tags
     */
    public StripMarkupTokenizer(String keepScript) {
        this.keepScript = Boolean.valueOf(keepScript);
    }

    /**
     * Eliminates tokens nested inside markup language tags; assumes that tokens have been split by line rather than using
     * whitespace
     */
    @Override
    public List<Token> tokenize(List<Token> tokens) throws SdTextException {
        List<Token> tokenVector = Collections.synchronizedList(new Vector<Token>());
        try {
            synchronized (tokenVector) {
                boolean script = false;
                for (Token token : tokens) {
                    for (String name : token.name().split(">")) {

                        // Printing error message instead of throwing exception because govdocs has unbalanced brackets
                        if (name.indexOf("<") != name.lastIndexOf("<")) {
                            logger.log(Level.SEVERE, "Unbalanced brackets error", name);
                        }
                        if (name.contains("script") && name.contains("<")) {
                            script = !script;
                        }

                        if (!script) {
                            if (name.indexOf('<') > 0) {
                                tokenVector.add(new Token(name.substring(0, name.indexOf('<')), token.start_byte(), token.end_byte()));
                            } else if (name.indexOf('<') < 0) {
                                tokenVector.add(new Token(name, token.start_byte(), token.end_byte()));
                            }
                        } else if (script && this.keepScript) {
                            tokenVector.add(new Token(name, token.start_byte(), token.end_byte()));
                        }
                    }
                }
                return tokenVector;
            }
        } catch (Exception e) {
            ExceptionHandler.handleException(e);
        }
        return tokenVector;
    }
}
