package edu.georgetown.gucs.utility;

import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.IOException;
import java.util.logging.Logger;


/**
 * Checks file existance and permissions
 *
 * @author Clay Shields
 */

public class FileChecker {


    /**
     *
     * Takes a String that should be the name of a readable file. Checks that it exists and is
     * readable, and returns a File object that references the named file if possible. Otherwise,
     * throws an SdTextException
     *
     * @param file A String containing a filename
     *
     * @param callerClass A String with the name of the calling class used to specify the exception
     *
     * @param callerMethod A String with the name of the calling method used to specify the exception
     *
     **/
    
    public static File canRead(String file, String callerClass, String callerMethod) throws SdTextException{

	if (file == null){
	    throw new SdTextException(callerClass + ":" + callerMethod + ":Null filename for reading");
	}
	
	File toCheck = new File(file);
	
	if (!toCheck.exists()){
	    throw new SdTextException(callerClass + ":" + callerMethod + ":File doesn't exist to read:" + file);
	}

	if (!toCheck.canRead()){
	    throw new SdTextException(callerClass + ":" + callerMethod + ":File unreadable:" + file);
	}

	return toCheck;
	
    }

    /**
     *
     * Takes a String that should be the name of a writeable file or directory. Checks that it exists and is
     * writeable, and returns a File object that references the named file or directory if possible. Otherwise,
     * throws an SdTextException
     *
     * @param file A String containing a filename
     *
     * @param callerClass A String with the name of the calling class used to specify the exception
     *
     * @param callerMethod A String with the name of the calling method used to specify the exception
     *
     **/
    
    public static File canWrite(String file, String callerClass, String callerMethod) throws SdTextException{

	File parent = null;
	
	if (file == null){
	    throw new SdTextException(callerClass + ":" + callerMethod + ":Null filename for writing");
	}

	File toCheck = new File(file);

	// File is either a specific file or a directory
	// Check if it is a directory first and if it is writable
	// If not, then check if it is a File and if it exists and is writable


	if (toCheck.isDirectory()){
	    if (!toCheck.canWrite()){
		throw new SdTextException(callerClass + ":" + callerMethod + ":Cannot write to directory:" + file);
	    }
	    // ok otherwise
	} else if (toCheck.isFile()){
	    if (!toCheck.canWrite()){
		throw new SdTextException(callerClass + ":" + callerMethod + ":Cannot write to file:" + file);
	    }
	    // ok otherwise
	}
	else {
	    // Not clear what we have. Perhaps a non-existant file on an existant path?
	    // Check the parent to see if it is a writeable directory
	    File parentPath = toCheck.getParentFile();
	    if (parent == null){
		// Ok, here we seem to have something that doesn't show a parent. On windows,
		// this seems to happen when just a filename is given for the current directory.
		// So lets move ahead and see what happens.
	    }  
	    else if (parent.isDirectory()){
		if (!parent.canWrite()){
		    throw new SdTextException(callerClass + ":" + callerMethod + ":Cannot write to directory:" + parent);
		}		
	    }
	}

	// If we get here it doesn't mean there won't be problems later, we just don't see them now.

	return toCheck;	
    }
}
    
  



