package sdtext;

import edu.georgetown.gucs.dictionary.Dictionary;
import edu.georgetown.gucs.fingerprinter.*;
import edu.georgetown.gucs.matcher.*;
import edu.georgetown.gucs.tokenizers.TokenizerList;
import edu.georgetown.gucs.utility.*;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.Serializable;
import java.util.*;

/**
 * 
 * This is the API defined for the C# interface
 * Created by Evan B.
 *
 * @since 3/17/15.
 */
public class API {

    private final SortedMap<String, Integer> scores = Collections.synchronizedSortedMap(new TreeMap<String, Integer>());
    /**
     * 1F_REQ5
     * Provides error handling feedback for debugging purposes
     */
    private static final org.slf4j.Logger logger = LoggerFactory.getLogger(API.class);

    /**
     * 1F_REQ1
     * Compare list of fingerprint files to a text document
     *
     * @param threshold            the minimum score required for a match
     * @param match                name of the matcher to use for this comparison
     * @param documentContent      text content of the document to be compared
     * @param fingerprintFileNames list of the fingerprint file names to compare with the given text document
     * @param dictionary           the dictionary to use to create the fingerprints
     * @param loggingLVL           the parameter which enables logging to record events
     * @return a mapthat contains the score,filename of the fingerprints compared to the text
     */
    public Results cmpFingerprintFilesToText(Integer threshold,
                                         String match,
                                         String documentContent,
                                         String dictionary,
                                         ArrayList<String> fingerprintFileNames,
                                         String loggingLVL)
	throws SdTextException {

	// Set logging level if specified
        if (loggingLVL != null)
            Global.setLoggingLvl(loggingLVL);

	try {

	    // Check the range of the threshold parameter
	    if (threshold < Global.MIN_SCORE || threshold > Global.MAX_SCORE){
		throw new SdTextException("API:cmpFingerprintFilesToText:Threshold score out of range");	    
	    }

	    // Check that the dictionary is ok	    
	    FileChecker.canRead(dictionary,"API","cmpFingerprintFilesToText");
	    Dictionary dict = new Dictionary(dictionary);

	    // See if we can create the matcher given the name we have
	    FingerprintMatcher matcher = (FingerprintMatcher) ObjectBuilder.buildObject(match);
	    
	
	    // Read fingerprints out of fingerprint files
	    // Check and make sure they are all readable

            ArrayList<Fingerprinter> fingerprintFiles = new ArrayList<>();
            for (String fingerprints : fingerprintFileNames) {
		FileChecker.canRead(fingerprints,"API","cmpFingerprintFilesToText");		
		Fingerprinter tempFP = new Fingerprinter(fingerprints, dictionary);
                fingerprintFiles.add(tempFP);
            }

	    // if we didn't get any files we have a problem
            if (fingerprintFiles.isEmpty()) {
		throw new SdTextException("API:cmpFingerprintFilesToText:List of fingerprint files appeared empty");
            }

	    // See if we can create the matcher given the name we extract from the first fingerprint
	    Fingerprinter fingerprinter = (Fingerprinter) ObjectBuilder.buildObject(fingerprintFiles.get(0).getFingerprintName());

            //get the tokenizers used to create the fingerprinted files
            TokenizerList tokenizers = dict.getTokenizers();
	            
            //delete first tokenizer and change to memory tokenizer
            //tokenize file with the list of tokenizers
            tokenizers.replaceFileTokenizer("MemoryTokenizer");

            List<Fingerprint> stringFingerprint = new ArrayList<>();

            fingerprinter.setDictionary(dict);
            stringFingerprint.addAll(fingerprinter.computeFingerprint(tokenizers, documentContent));


            //compare fingerprint to vector of fingerprints

            for (Fingerprinter fingTemp : fingerprintFiles) {

                int score = matcher.getScore(stringFingerprint.get(0).getFingerprint(), fingTemp.getBase64Fingerprints().get(0).getFingerprint());

                if (score >= threshold) {

                    scores.put(fingTemp.getFileName(), score);
                }
            }
        } catch (SdTextException sdtext) {
            throw sdtext;
        } catch (Exception e) {
            ExceptionHandler.handleException(e);
        }
        return new Results(scores);
    }

    /**
     * 1F_REQ2
     * Compare a fingerprint file to a directory
     *
     * @param threshold       minimum score required to be considered a match
     * @param match           name of the matcher to use for this comparison
     * @param fingerprintFile filename of the fingerprint file to compare with this directory
     * @param directory       directory path to compare with this fingerprint
     * @param dictionary      the dictionary to use to create the fingerprints
     * @param loggingLVL      the parameter which enables logging to record events
     * @return the file name and score that was compared to a fingerprint
     */
    public Results cmpFingerprintToDirectory(Integer threshold,
                                             String match,
                                             String fingerprintFile,
                                             String directory,
                                             String dictionary,
                                             String loggingLVL) throws SdTextException {
        if (loggingLVL != null)
            Global.setLoggingLvl(loggingLVL);

        ArrayList<Pair<String, Integer>> results = new ArrayList<>();
        try {

	    // All of the parameters are passed to CompareFingerprint and checked there
	    // so no need to error check here as well
	    
            CompareFingerprint compare = new CompareFingerprint(match, fingerprintFile, directory, threshold, dictionary);

            results.addAll(compare.compareFingerprintDirectory());
            Collections.sort(results, new HightoLowScore());

        } catch (SdTextException sdtext) {
            throw sdtext;
        } catch (Exception e) {
            ExceptionHandler.handleException(e);
        }
        return new Results(results);
    }


    /**
     * 1F_REQ3
     * Create a dictionary and normalize (trim) its IDF range (0-1]
     *
     * @param minIDF     minimum normalized IDF to keep in trimmed dictionary 0.0 (default 0.0)
     * @param maxIDF     maximum normalized IDF to keep in trimmed dictionary (default 1.0)
     * @param config     filename of the tokenizer configuration file
     * @param directory  directory containing a set of sample files
     * @param dictionary output dictionary file name
     * @param loggingLVL the parameter which enables logging to record events
     */
    public void createDictAndTrim(Double minIDF,
                                  Double maxIDF,
                                  String config,
                                  String directory,
                                  String dictionary,
                                  String loggingLVL) throws SdTextException {

	if (loggingLVL != null)
            Global.setLoggingLvl(loggingLVL);

	// Config, directory and dictionary are error checked in makeDictionary
	if (minIDF < 0 || minIDF > 1) {
	    throw new SdTextException("API:createDictAndTrim: minIDF out of range, must be between 0 and 1");
	}

	if (maxIDF < 0 || maxIDF > 1) {
	    throw new SdTextException("API:createDictAndTrim: maxIDF out of range, must be between 0 and 1");
	}
	
	
        try {
            Dictionary dict = new Dictionary();

	    // Make Dictionary checks all the input files, so pass them on without checking
	    dict.makeDictionary(config, directory, dictionary);
	    // Trim the dictionary
            dict.trimByIDF(minIDF, maxIDF);
	    // Output the dictionary
            dict.saveDictionaryXML(dictionary);
        } catch (SdTextException sdtext) {
            throw sdtext;
        } catch (Exception e) {
            ExceptionHandler.handleException(e);
        }
    }

    /**
     * 1F_REQ4
     * Create fingerprint for file
     * NOTE --- The fingerprinter is [default] BitVectorFingerprinter
     *
     * @param fingerprintFile the file to fingerprint
     * @param dictionary      the dictionary filename
     * @param outputFile      output fingerprint filename
     * @param loggingLVL      the parameter which enables logging to record events
     */
    public void createFingerprint(String fingerprintFile,
                                  String dictionary,
                                  String outputFile,
                                  String loggingLVL) throws SdTextException {

        if (loggingLVL != null)
            Global.setLoggingLvl(loggingLVL);

        try {
	    FileChecker.canRead(fingerprintFile, "API","createFingerprint");
	    FileChecker.canRead(dictionary, "API","createFingerprint");
	    FileChecker.canWrite(outputFile, "API","createFingerprint");
	    
            BitVectorFingerprinter bvfp = new BitVectorFingerprinter(dictionary);
            bvfp.setOutput(true, false, false);
            bvfp.generateXML(bvfp.computeFingerprint(fingerprintFile), dictionary, outputFile);


        } catch (SdTextException sdtext) {
            throw sdtext;
        } catch (Exception e) {
            ExceptionHandler.handleException(e);
        }
    }


    /**
     * Comparator that orders scores highest to lowest
     */
    private static class HightoLowScore implements Comparator<Pair<String, Integer>>, Serializable {

        @Override
        public int compare(Pair<String, Integer> o1, Pair<String, Integer> o2) {

            //by using the second argument we get descending order
            return o2.getB().compareTo(o1.getB());
        }
    }

}
